/* SPDX-License-Identifier: GPL-2.0+ WITH Linux-syscall-note */
/* Truncated header from the kernel sources that just defines the name sizes
 * below; used by icmp.h */
#ifndef _LINUX_IF_H
#define _LINUX_IF_H

#define	IFNAMSIZ	16
#define	IFALIASZ	256
#define	ALTIFNAMSIZ	128

#endif /* _LINUX_IF_H */
